from __future__ import absolute_import, division, print_function, with_statement

CLOSED = 'CLOSED'
AWAITING_AUTHENTICATION = 'AWAITING_AUTHENTICATION'
AWAITING_CONNECTION = 'AWAITING_CONNECTION'
AUTHENTICATING = 'AUTHENTICATING'
OPEN = 'OPEN'
ERROR = 'ERROR'
RECONNECTING = 'RECONNECTING'
CHALLENGING = 'CHALLENGING'
