CONNECTION_ERROR = 'connectionError'
CONNECTION_STATE_CHANGED = 'connectionStateChanged'
MAX_RECONNECTION_ATTEMPTS_REACHED = 'MAX_RECONNECTION_ATTEMPTS_REACHED'
CONNECTION_AUTHENTICATION_TIMEOUT = 'CONNECTION_AUTHENTICATION_TIMEOUT'
ACK_TIMEOUT = 'ACK_TIMEOUT'
NO_RPC_PROVIDER = 'NO_RPC_PROVIDER'
RESPONSE_TIMEOUT = 'RESPONSE_TIMEOUT'
DELETE_TIMEOUT = 'DELETE_TIMEOUT'
UNSOLICITED_MESSAGE = 'UNSOLICITED_MESSAGE'
MESSAGE_DENIED = 'MESSAGE_DENIED'
MESSAGE_PARSE_ERROR = 'MESSAGE_PARSE_ERROR'
VERSION_EXISTS = 'VERSION_EXISTS'
NOT_AUTHENTICATED = 'NOT_AUTHENTICATED'
MESSAGE_PERMISSION_ERROR = 'MESSAGE_PERMISSION_ERROR'
LISTENER_EXISTS = 'LISTENER_EXISTS'
NOT_LISTENING = 'NOT_LISTENING'
TOO_MANY_AUTH_ATTEMPTS = 'TOO_MANY_AUTH_ATTEMPTS'
IS_CLOSED = 'IS_CLOSED'
RECORD_NOT_FOUND = 'RECORD_NOT_FOUND'
NOT_SUBSCRIBED = 'NOT_SUBSCRIBED'
INVALID_RPC_CORRELATION_ID = 'INVALID_RPC_CORRELATION_ID'
